/*-
 * Copyright (C) @BABOLO  2008 Feb 10
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2008 Feb 10\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: db_pgsql.c,v 1.24 2009/10/27 02:01:14 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <poll.h>
#include <mife.h>
#include <err.h>
#include <libpq-fe.h>
#include <babolo/parser.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"

#define PRINTNOTNULL(A) (!(A) ? "" : (A))
#define PRINTANDNULL(A) PRINTNOTNULL(A), ((A) ? "~" : "")
#define PQSTATNDNULL(A) PRINTANDNULL(PQparameterStatus(db->conn, (A)))
#define ALIGN8(A)  popt[i].A, (strlen(popt[i].A) < 8 ? "		" : "	")
#define ALIGN16(A) popt[i].A                                                                          \
      , ((strlen(popt[i].A) < 8) ? "		" : (strlen(popt[i].envvar) < 16) ? "	" : " ")

#define PQEXEC(RESULT, QUERY, CONDITION, CCODE, RCONN) {                                              \
    ifBLIN_QV5(db->flags) fprintf(stderr, "PQexec: %s\n", (char*)(QUERY));                            \
    (RESULT) = PQexec(RCONN, (QUERY));                                                                \
    if  (!(RESULT)) {                                                                                 \
        ifBLIN_QV1(db->flags) warnx("result is absent - fatal error.\n");                             \
        return(CCODE);                                                                                \
    }                                                                                                 \
    ifBLIN_QV5(db->flags) fprintf(stderr, "PQresultStatus %d\n", PQresultStatus(RESULT));             \
    if  (PQresultStatus(RESULT) != (CONDITION)) {                                                     \
        if  (PQresultStatus(RESULT) == PGRES_FATAL_ERROR) {                                           \
            ifBLIN_QV1(db->flags)                                                                     \
                warnx("PQexec failed: %s", PQresultErrorMessage(RESULT));                             \
            return(CCODE);                                                                            \
        } else if ((CONDITION) == PGRES_COMMAND_OK && PQresultStatus(RESULT) == PGRES_TUPLES_OK) {    \
            ifBLIN_QV2(db->flags)                                                                     \
                warnx("PQexec wait for COMMAND_OK but received TUPLES_OK");                           \
        } else if ((CONDITION) == PGRES_COMMAND_OK && PQresultStatus(RESULT) == PGRES_EMPTY_QUERY) {  \
            ifBLIN_QV2(db->flags)                                                                     \
                warnx("PQexec wait for COMMAND_OK but received EMPTY_QUERY");                         \
        } else ifBLIN_QV1(db->flags) {                                                                \
            warnx("PQexec wait for %d but received %d", (CONDITION), PQresultStatus(RESULT));         \
            return(CCODE);                                                                            \
}   }   }

static void
NoticeProcessor(void *db, const char *message) {
    ifBLIN_QV1(((pgoblin_conn *)db)->flags) fprintf(stderr, "%s", message);
}

static int
finit(u_int32_t flags, pgoblin_dbases *this) {
    return(0);
}

static int
shurecon(pgoblin_conn *db) {
    PQconninfoOption *popt;
    int i, ex = EX_OK;
    PGresult *res;

    if  (db->conn != NULL && PQstatus(db->conn) == CONNECTION_OK) return(0);
    ifBLIN_QV5(db->flags) {
        fprintf( stderr
               , "Connection defaults:\n"
                 "size=chr keyword	envvar		compiled val		label\n"
               );
        popt = PQconndefaults();
        for (i = 0; popt[i].keyword; i++) {
            fprintf( stderr
                   , "%4d=%s~	%s%s%s%s%s%s%s%s%s\n"
                   , popt[i].dispsize
                   , popt[i].dispchar
                   , ALIGN8(keyword)
                   , ALIGN16(envvar)
                   , PRINTNOTNULL(popt[i].compiled)
                   , !(popt[i].compiled) ? "	" : strlen(popt[i].compiled) < 7 ? "~	" : "~ "
                   , PRINTNOTNULL(popt[i].val)
                   , !(popt[i].val)
                   ? "		"
                   : strlen(popt[i].val) < 7 ? "~		" : "~	"
                   , popt[i].label
                   );
        }
        PQconninfoFree(popt);
        fprintf( stderr
               , "PGREALM=%s%s PGDATESTYLE=%s%s PGTZ=%s%s PGCLIENTENCODING=%s%s PGGEQO=%s%s\n"
               , PRINTANDNULL(getenv("PGREALM"))
               , PRINTANDNULL(getenv("PGDATESTYLE"))
               , PRINTANDNULL(getenv("PGTZ"))
               , PRINTANDNULL(getenv("PGCLIENTENCODING"))
               , PRINTANDNULL(getenv("PGGEQO"))
               );
    }
    MARK_R_CONN_GO(0);;
    ;;  if  (db->flags & PGOBLIN_PARMIN && db->dbname) {
    ;;      db->conn = PQconnectdb(db->dbname);
    ;;  } else {
    ;;      db->conn = PQsetdbLogin(db->host, db->port, NULL, NULL, db->dbname, db->username, NULL);
    ;;  }
    ;;  db->flags |= PGOBLIN_CLOSECONN;
    MARK_R_CONN_WENT(0);;
    ifBLIN_QV5(db->flags) {
        fprintf(stderr, "PQoptions=%s~\n", PQoptions(db->conn));
        fprintf( stderr
               , "PQdb=%s~ PQuser=%s~ PQpass=%s%s "
                 "PQhost=%s%s PQport=%s%s "
                 "PQtty=%s%s\n"
               , PQdb(db->conn), PQuser(db->conn)
               , PRINTANDNULL(PQpass(db->conn))
               , PRINTANDNULL(PQhost(db->conn))
               , PRINTANDNULL(PQport(db->conn))
               , PRINTANDNULL(PQtty(db->conn))
               );
        fprintf( stderr
               , "server_version=%s%s "
                 "client_encoding=%s%s "
                 "is_superuser=%s%s "
                 "session_authorization=%s%s "
                 "DateStyle=%s%s\n"
               , PQSTATNDNULL("server_version")
               , PQSTATNDNULL("client_encoding")
               , PQSTATNDNULL("is_superuser")
               , PQSTATNDNULL("session_authorization")
               , PQSTATNDNULL("DateStyle")
               );
        fprintf( stderr
               , "PQprotocolVersion=%d PQsocket=%d PQbackendPID=%d\n"
               , PQprotocolVersion(db->conn)
               , PQsocket(db->conn)
               , PQbackendPID(db->conn)
               );
        fprintf(stderr, "PQstatus %d\n", PQstatus(db->conn));
    }
    PQsetNoticeProcessor(db->conn, NoticeProcessor, db);
    if  (PQstatus(db->conn) != CONNECTION_BAD) {
        if  (  db->flags & PGOBLIN_NEEDINTIME
            && !(db->flags & (PGOBLIN_INT64TIME | PGOBLIN_FLOATTIME))
            ) {
            if  (!db->intran) {
/*          if  (PQtransactionStatus(db->conn) != PQTRANS_INTRANS) { */
                PQEXEC(res, "BEGIN", PGRES_COMMAND_OK, -1, db->conn);
                PQclear(res);
            }
            PQEXEC( res
                  , "DECLARE pgoblin BINARY NO SCROLL CURSOR FOR "
                    "SELECT '2000-01-01 00:00:01'::timestamp;"
                    "FETCH 1 FROM pgoblin"
                  , PGRES_TUPLES_OK, -1
                  , db->conn
                  );
            if  (PQgetisnull(res, 0, 0) || (PQgetlength(res, 0, 0) < 8)) {
                ex = EX_DATAERR;
                errno = EINVAL;
            } else {
                if  (*((u_char*)(PQgetvalue(res, 0, 0))) == 077) {
                    db->flags |= PGOBLIN_FLOATTIME;
                    ifBLIN_QV3(db->flags) fprintf(stderr, "shurecon float timestamp format\n");
                } else if (*((u_char*)(PQgetvalue(res, 0, 0))) == 0) {
                    db->flags |= PGOBLIN_INT64TIME;
                    ifBLIN_QV3(db->flags) fprintf(stderr, "shurecon int64 timestamp format\n");
                } else {
                    ifBLIN_QV1(db->flags) fprintf(stderr, "shurecon unknown timestamp format\n");
                    ex = EX_DATAERR;
                    errno = EFTYPE;
            }   }
            PQclear(res);
            PQEXEC(res, "CLOSE pgoblin", PGRES_COMMAND_OK, -1, db->conn);
            PQclear(res);
            if  (!db->intran) {
                PQEXEC(res, "END", PGRES_COMMAND_OK, -1, db->conn);
                PQclear(res);
        }   }
        return(0);
    }
    ifBLIN_QV1(db->flags) warnx("Connection to PostgreSQL failed: %s\n", PQerrorMessage(db->conn));
    return(1);
}

static int
tarry(void *connv, int timeout) {
    PGconn *conn = (PGconn *)connv;
    int           sock, i = 0;
    struct pollfd pfd;

    if  ((sock = PQsocket(conn)) < 0) {
        i = -2;
        errno = EINVAL;
    } else if (timeout != 0) {
        bzero(&pfd, sizeof(pfd));
        pfd.fd = sock;
        pfd.events = POLLIN | POLLRDNORM | POLLRDBAND | POLLPRI;
        i = poll(&pfd, 1, timeout);
    }
    if  (i >= 0) {
        if  (!PQconsumeInput(conn)) i = -1;
    }
    return(i);
}

#define U0 0
#define U1 1
#define U2 2
#define U3 3
#define U4 4

static u_char miss[][PGOBLIN_CR_MAX] =
/*  OK   TUPL OUT  IN   STR  BIN   */
{{  U1,  U1,  U1,  U1,  U2,  U2 } /* PGRES_EMPTY_QUERY    */
,{  U0,  U2,  U2,  U2,  U2,  U2 } /* PGRES_COMMAND_OK     */
,{  U1,  U0,  U2,  U2,  U2,  U2 } /* PGRES_TUPLES_OK      */
,{  U2,  U2,  U0,  U2,  U2,  U2 } /* PGRES_COPY_OUT       */
,{  U2,  U2,  U2,  U0,  U2,  U2 } /* PGRES_COPY_IN        */
,{  U3,  U3,  U3,  U3,  U3,  U3 } /* PGRES_BAD_RESPONSE   */
,{  U3,  U3,  U3,  U3,  U3,  U3 } /* PGRES_NONFATAL_ERROR */
,{  U4,  U4,  U4,  U4,  U4,  U4 } /* PGRES_FATAL_ERROR    */
};

static ExecStatusType es[] = {PGRES_COMMAND_OK, PGRES_TUPLES_OK, PGRES_COPY_OUT, PGRES_COPY_IN};

static int
extatus(u_int32_t flags, PGresult *res, const char *who, u_char cond) {
    ExecStatusType st;
    int ex = 0;

    st = PQresultStatus(res);
    ifBLIN_QV5(flags)
        fprintf(stderr, "PQresultStatus %d %s, diff %d\n", st, PQresStatus(st), miss[st][cond]);
    switch (miss[st][cond]) {
    case U1:
        ifBLIN_QV2(flags) warnx( "%s waits for %d %s but received %d %s"
                               , who, es[cond], PQresStatus(es[cond])
                               , st, PQresStatus(st)
                               );
        break;
    case U2:
        ifBLIN_QV1(flags) warnx( "%s wait for %d %s but received %d %s"
                               , who, es[cond], PQresStatus(es[cond])
                               , st, PQresStatus(st)
                               );
        ex = miss[st][cond];
        break;
    case U4:
    case U3:
        ifBLIN_QV1(flags) warnx("%s failed: %s", who, PQresultErrorMessage(res));
        ex = miss[st][cond];
        break;
    }
    return(ex);
}

static int
doquery(pgoblin_conn *db, void **rsv, const char *query, u_char cond) {
    PGresult *res, **rs = (PGresult **)rsv;
    char *q, *p;
    int ex = 0;

    res = NULL;
    switch (cond) {
    case PGOBLIN_CR_OK:
    case PGOBLIN_CR_TUPL:
    case PGOBLIN_CR_OUT:
    case PGOBLIN_CR_IN:
        ifBLIN_QV5(db->flags) fprintf(stderr, "PQexec %d=%s~\n", cond, query);
        if  (!(res = PQexec(db->conn, query))) {
            ifBLIN_QV1(db->flags)
                warnx("PQ result is absent - fatal error %s", PQerrorMessage(db->conn));
            ERROUT(EX_UNAVAILABLE, ENODEV);
        }
        ex = extatus(db->flags, res, "PQexec", cond);
        break;
    case PGOBLIN_CR_SSEL:
        p = "pgoblin";
        goto rest;
    case PGOBLIN_CR_BSEL:
        p = "pgoblin BINARY";
    rest:
        if  (!db->intran && (ex = doquery(db, NULL, "BEGIN", PGOBLIN_CR_OK))) goto out;
        asprintf(&q, "DECLARE %s NO SCROLL CURSOR FOR %s", p, query);
        if  (!q) {
            ifBLIN_QV1(db->flags) warnx("doquery: No memory");
            ERROUT(EX_UNAVAILABLE, ENOMEM);
        }
        ex = doquery(db, (void**)&res, q, PGOBLIN_CR_OK);
        free(q);
        break;
    default:
        ifBLIN_QV1(db->flags) warnx("doquery pgsql illegal cond=%d", cond);
        ex = -1;
    }
out:
    if  (rs) *rs = res; else if (res) PQclear(res); /* XXXX */
    return(ex);
}

static int
prepare(pgoblin_conn *db, void **rs, const char *query) {
    PGresult *res;
    int ex = 0;

    res = NULL;
    ifBLIN_QV5(db->flags) fprintf(stderr, "PQprepare =%s~\n", query);
    if  (!(res = PQprepare(db->conn, "", query, 0, NULL))) {
        ifBLIN_QV1(db->flags) warnx("PQ result is absent - fatal error %s", PQerrorMessage(db->conn));
        ERROUT(EX_UNAVAILABLE, ENODEV);
    }
out:
    if  (res) PQclear(res);
    if  (!ex) *rs = "";
    return(ex);
}

static int
execute( pgoblin_conn *db
       , void **rsv
       , const void *qp, int cnt, const char * const *val, const int *len
       , u_char cond
       ) {
    PGresult **rs = (PGresult**)rsv;
    const char *query = (char *)qp;
    char const *kind;
    PGresult *res;
    int ex = 0;

    res = NULL;
    kind = (cond & PGOBLIN_ALTERQ) ? "PQexecPrepared" : "PQexecParams";
    switch (cond & ~PGOBLIN_ALTERQ) {
    case PGOBLIN_CR_OK:
    case PGOBLIN_CR_TUPL:
        ifBLIN_QV5(db->flags) fprintf(stderr, "%s %d=%s~\n", kind, cond, query);
        if  (!(res = (cond & PGOBLIN_ALTERQ)
                   ? PQexecPrepared(db->conn, query, cnt,       val, len, NULL, 0)
                   : PQexecParams(  db->conn, query, cnt, NULL, val, len, NULL, 0)
            ) ) {
            ifBLIN_QV1(db->flags)
                warnx("PQ result is absent - fatal error %s", PQerrorMessage(db->conn));
            ERROUT(EX_UNAVAILABLE, ENODEV);
        }
        ex = extatus(db->flags, res, kind, cond);
        break;
    default:
        ifBLIN_QV1(db->flags) warnx("execute pgsql illegal cond=%d", cond);
        ex = -1;
    }
out:
    if  (rs) *rs = res; else if (res) PQclear(res);
    return(ex);
}

static char *
notifies(void *connv) {
    PGconn *conn = (PGconn *)connv;
    PGnotify *n;
    char *c = NULL;

    n = PQnotifies(conn);
    if  (n) {
        if  (n->relname) c = strdup(n->relname);
        PQfreemem(n);
    }
    return(c);
}

static int
p1info(int kind, void *resv) {
    PGresult *res = (PGresult *)resv;
    char *c;
    int ex;

    switch (kind) {
    case PGOBLIN_Ntuples: ex =   PQntuples(res); break;
    case PGOBLIN_Nfields: ex =   PQnfields(res); break;
    case PGOBLIN_Nchange:
        c = PQcmdTuples(res);
        ex = *c ? strtoul(c, NULL, 10) : -1;
        break;
    case PGOBLIN_LastOid: ex =  PQoidValue(res); break;
    default             : ex = -1;
    }
    return(ex);
}

static char *
p2info(int kind, void *resv, int sub) {
    PGresult *res = (PGresult *)resv;
    char *cx;

    switch (kind) {
    case PGOBLIN_ColName: cx =     PQfname(res, sub); break;
    case PGOBLIN_CmdStat: cx = PQcmdStatus(res     ); break;
    default             : cx = NULL;
    }
    return(cx);
}

static int
p3info(int kind, void *resv, int row, int col) {
    PGresult *res = (PGresult *)resv;
    int ex;

    switch (kind) {
    case PGOBLIN_IsNull : ex = PQgetisnull(res, row, col); break;
    case PGOBLIN_Length : ex = PQgetlength(res, row, col); break;
    case PGOBLIN_BinForm: ex =   PQfformat(res,      col); break;
    case PGOBLIN_TypeOid: ex =     PQftype(res,      col); break;
    case PGOBLIN_TypeMod: ex =      PQfmod(res,      col); break;
    default             : ex = -1;
    }
    return(ex);
}

static int
finish(u_int32_t flags, int kind, void **entity) {
    int ex = EX_OK;

    switch (kind) {
    case PGOBLIN_Connect: {
            PGconn **conn = (PGconn **)entity;

            if  (conn && *conn) {
                PQfinish(*conn);
                *conn = NULL;
        }   }
        break;
    case PGOBLIN_PFormed: {
            char **stmt = (char **)entity;

            if  (stmt && *stmt) {
                if  (**stmt) /* XXXX */;
                *stmt = NULL;
        }   }
        break;
    case PGOBLIN_FrCOPY:
    case PGOBLIN_FrNote:
        if  (entity && *entity) {
            PQfreemem(*entity);
            *entity = NULL;
        }
        break;
    case PGOBLIN_CleaRes: {
        PGresult **res = (PGresult **)entity;
            if  (res && *res) {
                PQclear(*res);
                *res = NULL;
        }   }
        break;
    default:
        ex = EX_UNAVAILABLE;
        errno = ENOTSUP;
    }
    return(ex);
}

static int
getstream(pgoblin_conn *db, void **resv, int quant, u_char cond) {
    PGresult **res = (PGresult**)resv;
    int ex = EX_OK;
    char *query;

    *res = NULL;
    if  ((cond != PGOBLIN_CR_SSEL) && (cond != PGOBLIN_CR_BSEL)) {
        ifBLIN_QV1(db->flags) warnx("getstream: Illegal cond %d", cond);
        ERROUT(EX_USAGE, EINVAL);
    }
    if  (asprintf(&query, "FETCH %d FROM pgoblin", quant) < 0) {
        ifBLIN_QV1(db->flags) warnx("getstream: No memory");
        ERROUT(EX_OSERR, ENOMEM);
    }
    ex = doquery(db, (void**)res, query, PGOBLIN_CR_TUPL);
    free(query);
out:
    return(ex);
}

static int
getcopy(pgoblin_conn *db, void **buf) {
    int ex;

    ex = PQgetCopyData(db->conn, (char**)buf, 0);
    return((ex < 0) ? (ex + 1) : ex);
}

static int
putcopy(pgoblin_conn *db, const void *buf, int nby) {
    return(PQputCopyData(db->conn, buf, nby));
}

static int
endstream(pgoblin_conn *db, u_char cond) {
    ExecStatusType st;
    PGresult *res;
    int ex = EX_OK;

    switch (cond) {
    case PGOBLIN_CR_IN:
        if  (PQputCopyEnd(db->conn, NULL) < 1) {
            ifBLIN_QV1(db->flags)
                warnx("endstream: endcopy failed: %s", PQerrorMessage(db->conn));
            ERROUT(-EX_DATAERR, EINVAL);
        }
        /* FALLTHRU */
    case PGOBLIN_CR_OUT:
        while ((res = PQgetResult(db->conn))) {
            st = PQresultStatus(res);
            ifBLIN_QV5(db->flags) fprintf(stderr, "PQresultStatus %d %s\n", st, PQresStatus(st));
            if  (st != PGRES_COMMAND_OK) {
                ifBLIN_QV1(db->flags) warnx("endstream pgsql: %s", PQresultErrorMessage(res));
                ex = EX_SOFTWARE;
                break;
            }
            PQclear(res);
            res = NULL;
        }
        break;
    case PGOBLIN_CR_SSEL:
    case PGOBLIN_CR_BSEL:
        ex = doquery(db, NULL, "CLOSE pgoblin", PGOBLIN_CR_OK);
        if  (!db->intran) {
            ex |= doquery(db, (void**)&res, "END", PGOBLIN_CR_OK);
            PQclear(res);
        }
        break;
    default:
        ifBLIN_QV1(db->flags) warnx("endstream pgsql %d", cond);
        ex = EX_USAGE;
    }
out:
    if  (ex) errno = EINVAL;
    return(ex);
}

pgoblin_dbases pgoblin_db_pgsql =
{ "3pGoblin-" VERS
, "pgsql\0" VERS
,                             finit         /* init      */
,                             finit         /* fini      */
,                             shurecon      /* shurecon  */
,                             doquery       /* query     */
,                             prepare       /* prepare   */
,                             execute       /* execute   */
,                             p1info        /* resinfo   */
,                             p2info        /* subinfo   */
,                             p3info        /* valinfo   */
, (char*(*)(void*, int, int)) PQgetvalue    /* getvalue  */
,                             getstream     /* getstream */
,                             getcopy       /* getcopy   */
,                             putcopy       /* putcopy   */
,                             endstream     /* endstream */
,                             tarry         /* tarry     */
,                             notifies      /* notifies  */
,                             finish        /* finish    */
};
